from datalogue.models.transformations.commons import Transformation
from datalogue.dtl_utils import _parse_string_list, _parse_list
from datalogue.errors import DtlError, _property_not_found
from typing import List, Union


class PathWithIndex(object):
    """
    Represents the associated tuple of a path with an index
    Used when we want to specify fields with a specific order.
    For example, the SplitByDelimiter transformation uses this class to specify the order of output fields
    according to their indices. If a record in the source has a value of 'a','b','c',
    then index 0 would refer to 'a', index 1 to 'b', etc.
    """

    type_str = "PathWithIndex"

    def __init__(self, index: int, path: List[str]):
        self.index = index
        self.path = path

    def __repr__(self):
        return f"PathWithIndex(index= {self.index}, path= {','.join(self.path)})"

    def __eq__(self, other: "PathWithIndex"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        return {"index": self.index, "field": self.path}

    @classmethod
    def _from_payload(cls, json: dict) -> Union[DtlError, "PathWithIndex"]:
        index = json.get("index")
        path = json.get("field")
        if path is None:
            return _property_not_found("field", json)
        if index is None:
            return _property_not_found("index", json)
        return cls(index=index, path=path)


class SplitByDelimiter(Transformation):
    """
    Splits data from one field (source) into multiple fields (target) by delimiter
    """

    type_str = "SplitByDelimiter"

    def __init__(self, path: List[str], delimiter: str, output_paths: List[PathWithIndex]):
        """
        :param path: path in the source data which contains data that user wants to split
        :param delimiter: delimiter used in the source, based on which the user would like to split the data into
        separate fields
        :param output_paths: ordered path for output with indices. After the split is performed on `path`,
        the output array will be assigned to the fields according to their indices.
        """
        Transformation.__init__(self, SplitByDelimiter.type_str)
        self.path = path
        self.delimiter = delimiter
        self.output_paths = output_paths

    def __eq__(self, other: "SplitByDelimiter"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"SplitByDelimiter(path= {self.path}, delimiter= {self.delimiter}, output_paths= {self.output_paths})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["sourceField"] = self.path
        base["delimiter"] = self.delimiter
        base["targetFields"] = list(map(lambda x: x._as_payload(), self.output_paths))
        return base

    @classmethod
    def _from_payload(cls, json: dict) -> Union[DtlError, "SplitByDelimiter"]:
        path = json.get("sourceField")
        delimiter = json.get("delimiter")
        indexed_paths = json.get("targetFields")

        if path is None:
            return _property_not_found("sourceField", json)
        if indexed_paths is None:
            return _property_not_found("targetFields", json)
        if delimiter is None:
            return _property_not_found("delimiter", json)

        output_paths = _parse_list(PathWithIndex._from_payload)(indexed_paths)
        return cls(path=path, delimiter=delimiter, output_paths=output_paths)



from uuid import UUID
from typing import List, Optional, Set, Tuple, Union
# done for backward compatability with 3.7
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from datalogue.clients._http import _HttpClient, HttpMethod
from datalogue.errors import DtlError
from datalogue.models.job import Job
from datalogue.models.pipeline import Pipeline
from datalogue.models.tag import Tag
from datalogue.dtl_utils import _parse_list
from datalogue.models.datastore import Datastore, _datastore_from_payload


class _TagClient:
    """
    Client to interact with the Tags
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    path_fragments = {"datastore": "datastores", "pipeline": "streams"}
    mappings = {
        "datastore": _datastore_from_payload,
        "pipeline": Pipeline._from_payload,
    }
    ENTITY_TYPE = Literal["datastore", "pipeline"]

    def update(self, old_name: str, name: str) -> Union[DtlError, Tag]:
        """
        Updates a tag

        :param id: id of the tag to be updated
        :param name: the updated name to be applied
        :return: Returns Tag object if successful, or DtlError if failed
        """
        searched_tag_rsp = self.search(old_name)

        if isinstance(searched_tag_rsp, DtlError):
            return searched_tag_rsp

        payload = {"name": name}
        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags/{searched_tag_rsp.id}", HttpMethod.PUT, payload
        )

        if isinstance(res, DtlError):
            return res

        return Tag._from_payload(res)

    def search(self, name: str) -> Union[DtlError, Tag]:
        """
        Retrieve a tag by its name.

        :param name: the name of the tag to be searched
        :return: the tag if successful, or a DtlError if failed
        """
        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags/{name}/search", HttpMethod.GET
        )

        if isinstance(res, DtlError):
            return res

        return Tag._from_payload(res)

    def prefix_search(self, prefix: str) -> Union[DtlError, List[Tag]]:
        """
        Search tags by its prefix.

        :param name: the name of the tag to be searched
        :return: the tag if successful, or a DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags/search",
            HttpMethod.GET,
            params={"starts_with": prefix},
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(Tag._from_payload)(res["tags"])

    def delete(self, name: str) -> Union[DtlError, bool]:
        """
        Delete a tag.

        :param name: the name of the tag to be deleted
        :return: Returns True if successful, or DtlError if failed
        """
        searched_tag_rsp = self.search(name)

        if isinstance(searched_tag_rsp, DtlError):
            return True

        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags/{searched_tag_rsp.id}", HttpMethod.DELETE
        )

        if isinstance(res, DtlError):
            return res

        return True

    def list(
        self, page: int = 1, item_per_page: int = 25
    ) -> Union[DtlError, List[Tag]]:
        """
        List all tags.

        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available tags if successful, or DtlError if failed
        """
        res = self.http_client.make_authed_request(
            self.service_uri + f"/tags?page={page}&size={item_per_page}", HttpMethod.GET
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(Tag._from_payload)(res)

    def _get_path_fragment(self, type: ENTITY_TYPE) -> Union[DtlError, str]:
        path_fragment = self.path_fragments.get(type)
        if path_fragment is None:
            return DtlError(f"Invalid entity type: {type}")
        return path_fragment

    def add_tag(
        self, entity_id: Union[str, UUID], tag_name: str, type: ENTITY_TYPE
    ) -> Union[DtlError, Datastore, Pipeline]:
        """
        Add tags to datastore.
        :param entity_id: id of entity to update
        :param tag_name: list of tag ids to add to the specified datastore
        :param type: type of the entity
        :return: Returns a datastore object if successful, or DtlError if failed
        """

        fragment = self._get_path_fragment(type)
        if isinstance(fragment, DtlError):
            return DtlError(f"Invalid entity type: {type}")

        res = self.http_client.make_authed_request(
            self.service_uri + f"/{fragment}/{entity_id}/tags",
            body=[tag_name],
            method=HttpMethod.PUT,
        )

        if isinstance(res, DtlError):
            return res
        return self.mappings[type](res)

    def remove_tag(
        self, entity_id: Union[str, UUID], tag_name: str, type: ENTITY_TYPE
    ) -> Union[DtlError, Datastore, Pipeline]:
        """
        Remove tags from datastore.
        :param entity_id: id of entity to update
        :param tag_name: list of tag ids to remove from the specified datastore
        :param type: type of the entity
        :return: Returns a datastore object if successful, or DtlError if failed
        """

        fragment = self._get_path_fragment(type)
        if isinstance(fragment, DtlError):
            return DtlError(f"Invalid entity type: {type}")

        res = self.http_client.make_authed_request(
            self.service_uri + f"/{fragment}/{entity_id}/tags",
            body=[tag_name],
            method=HttpMethod.DELETE,
        )

        if isinstance(res, DtlError):
            return res
        return self.mappings[type](res)



from typing import Union
from datalogue.dtl_utils import SerializableStringEnum

from datalogue.errors import _enum_parse_error, DtlError


class ComparisonOperator(SerializableStringEnum):
    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("operator", s)

    @staticmethod
    def _from_payload(json: str) -> Union[DtlError, "ComparisonOperator"]:
        return SerializableStringEnum.from_str(ComparisonOperator)(json)

    @staticmethod
    def from_str(string: str) -> Union[DtlError, "ComparisonOperator"]:
        return ComparisonOperator._from_payload(string)

    LT = "<"
    LTE = "<="
    EQ = "=="
    GT = ">"
    GTE = ">="
    NE = "!="

    # aliases, from the old class "CompareOperators"
    Eq = "=="
    Heq = ">="
    Leq = "<="
    Les = "<"
    Hig = ">"
    Neq = "!="



from typing import List, Union
from uuid import UUID

from datalogue.auth.authentication import (
    Authentications,
    _authentication_ref_from_payload,
)
from datalogue.clients._http import _HttpClient, HttpMethod
from datalogue.dtl_utils import _parse_list
from datalogue.errors import DtlError


class _AuthenticationClient:
    """
    Client to interact with the Authentication Schemes
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = ""

    def list(
        self, page: int = 1, item_per_page: int = 25
    ) -> Union[DtlError, List[Authentications]]:
        """
        Lists all the Authentication schemes defined for the users Organization

        :param page: Optional page number, default 1
        :param item_per_page: Optional items per page, default 25
        :return: A list of all the authentication schemes defined for the users organization
        """
        res = self.http_client.make_authed_request(
            self.service_uri
            + f"/user-provider-configs/public?page={page}&size={item_per_page}",
            HttpMethod.GET,
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(_authentication_ref_from_payload)(res)

    # todo - this method should return AuthenticationReference object
    def update(
        self, authentication_id: UUID, authentication_def: Authentications
    ) -> Union[DtlError, bool]:
        """
        Updates the given authentication definition

        :param authentication_id: UUID of the Authentication being updated
        :param authentication_def: Updated Authentication definition
        :return:
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/user-provider-configs/" + str(authentication_id),
            HttpMethod.PUT,
            authentication_def._as_payload(),
        )
        if isinstance(res, DtlError):
            return res
        return True

    def delete(self, authentication_id: UUID) -> Union[DtlError, bool]:
        """
        Delete an authentication definition

        :param authentication_id: UUID of the authentication to be deleted
        :return: DtlError if unsuccessful, or True is successful
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/user-provider-configs/" + str(authentication_id),
            HttpMethod.DELETE,
        )
        if isinstance(res, DtlError):
            return res
        return True



from uuid import UUID
from typing import List

from datalogue.models.permission import (
    Permission,
    SharePermission,
    UnsharePermission,
    ObjectType,
    Scope,
)
from datalogue.dtl_utils import is_valid_uuid
from datalogue.clients._http import _HttpClient, Union, HttpMethod, Optional
from datalogue.dtl_utils import _parse_list
from datalogue.errors import DtlError
from datalogue.models.data_product import DataProduct


class DataProductClient:
    """
    Client to interact with the Data Products
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, data_product: DataProduct) -> Union[DtlError, DataProduct]:
        """
        Creates a Data Product

        :param data_product: A Data Product object that user wants to create

        Restrictions:
        * data product naming must be unique

        :return: Returns created Data Product object if successful, or DtlError if failed
        """
        payload = data_product._as_payload()

        if payload["name"] == "":
            return DtlError("Provide a non-empty string as a name.")

        res = self.http_client.make_authed_request(
            self.service_uri + "/data-products", HttpMethod.POST, payload
        )

        if isinstance(res, DtlError):
            return res

        return DataProduct._from_payload(res)

    def update(
        self, id: UUID, name: Optional[str] = None, description: Optional[str] = None
    ) -> Union[DtlError, DataProduct]:
        """
        Updates a Data Product

        :param id: id of the Data Product to be updated
        :param name: New name to be applied to the data product (if 'name' is not being updated, `description` should be updated)
        :param description: New description to be applied to the data product (if 'description' is not being updated, `name` should be updated)
        :return: Returns an updated Data Product object if successful, or DtlError if failed
        """
        payload = {}

        if is_valid_uuid(id) is False:
            return DtlError("id provided is not a valid UUID format.")

        if name is None and description is None:
            return DtlError(
                "Either name or description must be mentioned to update a Data Product"
            )

        if name is not None:
            payload["name"] = name

        if description is not None:
            payload["description"] = description

        res = self.http_client.make_authed_request(
            self.service_uri + f"/data-products/{id}", HttpMethod.PUT, payload
        )

        if isinstance(res, DtlError):
            return res

        return DataProduct._from_payload(res)

    def get(self, id: UUID) -> Union[DtlError, DataProduct]:
        """
        Retrieve a Data Product by its id.

        :param id: id of an existing Data Product to be retrieved
        :return: Data Product object if successful, or DtlError if failed
        """
        if is_valid_uuid(id) is False:
            return DtlError("id provided is not a valid UUID format.")
        res = self.http_client.make_authed_request(
            self.service_uri + f"/data-products/{id}", HttpMethod.GET
        )
        if isinstance(res, DtlError):
            return res
        return DataProduct._from_payload(res)

    def add_pipelines(
        self, data_product_id: UUID, pipeline_ids: List[UUID]
    ) -> Union[DtlError, DataProduct]:
        """
        Add Pipelines to a Data Product.

        :param data_product_id: id of the Data Product
        :param pipeline_ids: ids of the pipelines that need to be added to a Data Product

        :Restrictions:
        * a pipeline can belong only to one data product at a time

        :return: updated Data Product object if successful, or DtlError if failed
        """
        if is_valid_uuid(data_product_id) is False:
            return DtlError("data_product_id provided is not a valid UUID format.")
        if len(pipeline_ids) == 0:
            return DtlError("Please specify at least 1 pipeline id under pipeline_ids")
        else:
            payload = {
                "streamIds": [str(pipeline_id) for pipeline_id in pipeline_ids],
                "streamAction": "add",
            }

        res = self.http_client.make_authed_request(
            self.service_uri + f"/data-products/{data_product_id}/streams",
            HttpMethod.POST,
            payload,
        )
        if isinstance(res, DtlError):
            return res
        return DataProduct._from_payload(res)

    def remove_pipelines(
        self, data_product_id: UUID, pipeline_ids: List[UUID]
    ) -> Union[DtlError, DataProduct]:
        """
        Remove Pipelines from a Data Product.

        :param data_product_id: id of the Data Product
        :param pipeline_ids: ids of the pipelines to be removed from a Data Product
        :return: updated Data Product object if successful, or DtlError if failed
        """
        if is_valid_uuid(data_product_id) is False:
            return DtlError("data_product_id provided is not a valid UUID format.")
        if len(pipeline_ids) == 0:
            return DtlError("Please specify at least 1 pipeline id under pipeline_ids")
        else:
            payload = {
                "streamIds": [str(pipeline_id) for pipeline_id in pipeline_ids],
                "streamAction": "remove",
            }

        res = self.http_client.make_authed_request(
            self.service_uri + f"/data-products/{data_product_id}/streams",
            HttpMethod.POST,
            payload,
        )
        if isinstance(res, DtlError):
            return res
        return DataProduct._from_payload(res)

    def delete(
        self, data_product_id: UUID, delete_pipelines: bool = True
    ) -> Union[DtlError, bool]:
        """
        Deletes the given Data Product

        :param data_product_id: id of the data product to be deleted
        :param delete_pipelines: cascade delete all pipelines with permissions of this Data Product, default True
        :return: true if successful, DtlError otherwise
        """
        if is_valid_uuid(data_product_id) is False:
            return DtlError("id provided is not a valid UUID format.")
        res = self.http_client.make_authed_request(
            self.service_uri
            + f"/data-products/{data_product_id}?delete-streams={delete_pipelines}",
            HttpMethod.DELETE,
        )
        if isinstance(res, DtlError):
            return res
        else:
            return True

    def list(
        self, by_name: str = "", page: int = 1, size: int = 25
    ) -> Union[DtlError, List[DataProduct]]:
        """
        List all the data products that are saved

        :param by_name: optionally filter your list by a data product name, containing the supplied keyword, default ''
        :param page: page to be retrieved, default 1
        :param size: number of items to be put in a page, default 25
        :return: Returns a List of all the available data products or DtlError
        """
        if not isinstance(by_name, str):
            return DtlError("list by_name accepts only one name, given as a string")

        params = {"page": page, "size": size, "search-in-name": by_name}
        res = self.http_client.make_authed_request(
            path=self.service_uri + f"/data-products",
            method=HttpMethod.GET,
            params=params,
        )
        if isinstance(res, DtlError):
            return res
        return _parse_list(DataProduct._from_payload)(res)

    def share(
        self,
        data_product_id: UUID,
        target_type: Scope,
        target_id: UUID,
        permission: Permission,
        share_all_pipelines: bool = True,
    ) -> Union[SharePermission, DtlError]:
        """
        Share the Data Product, and optionally its pipelines.

        :param data_product_id: id of the Data Product that is to be shared
        :param target_type: Scope (`User`, `Group`) with whom you want to share the template. It can be User, or Group.
        :param target_id: id of the User/Group you want to share with
        :param permission: Permission (`Read`/`Write`/`Share`) to be shared for the Data Product.
        :param share_all_pipelines: cascade permission sharing to all pipelines of this Data Product,
        that are accessible by the requesting user, default True
        :return: Returns a SharePermission for the specific Data Product if successful, or DtlError if failed
        """
        if isinstance(target_type, str) and target_type not in Scope._value2member_map_:
            return DtlError(
                f"target_type: {target_type} is invalid. It can be `User` or `Group`"
            )
        if (
            isinstance(permission, str)
            and permission not in Permission._value2member_map_
        ):
            return DtlError(
                f"permission: {permission} is invalid. It can be `Read`, `Write` or `Share`"
            )
        params = {
            "targetType": target_type.value,
            "targetId": f"{target_id}",
            "permission": permission.value,
            "update-all-stream-permissions": share_all_pipelines,
        }
        rsp = self.http_client.make_authed_request(
            path=f"{self.service_uri}/data-products/{str(data_product_id)}/shares",
            method=HttpMethod.POST,
            params=params,
        )

        if isinstance(rsp, DtlError):
            return rsp

        return SharePermission(
            ObjectType.DataProduct, target_id, target_type, permission
        )

    def unshare(
        self,
        data_product_id: UUID,
        target_type: Scope,
        target_id: UUID,
        permission: Permission,
        unshare_all_pipelines: bool = True,
        exclude_pipeline_ids: Optional[List[UUID]] = None,
    ) -> Union[DtlError, UnsharePermission]:
        """
        Unshare the Data Product, and/or all of its accessible pipelines
        (excluding pipeline_ids supplied by the requesting user).

        :param data_product_id: id of the Data Product that is to be unshared
        :param target_type: Scope (`User`, `Group`) you want to unshare the Data Product with. It can be User, or Group.
        :param target_id: id of the User/Group you want to unshare with
        :param permission: existing permission (`Read`/`Write`/`Share`) of the Data Product, that you want to revoke
        :param unshare_all_pipelines: cascade permission unsharing to all pipelines of Data Product, that
        are accessible by the requesting user (except exclude_pipeline_ids), default True
        :param exclude_pipeline_ids: A list of pipeline_ids, that will not be unshared, while unsharing Data Product
        :return: Returns a UnsharePermission for the specific Data Product if successful, or DtlError if failed
        """

        if isinstance(target_type, str) and target_type not in Scope._value2member_map_:
            return DtlError(
                f"target_type: {target_type} is invalid. It can be `User` or `Group`"
            )
        if (
            isinstance(permission, str)
            and permission not in Permission._value2member_map_
        ):
            return DtlError(
                f"permission: {permission} is invalid. It can be `Read`, `Write` or `Share`"
            )
        params = {
            "targetType": target_type.value,
            "targetId": f"{target_id}",
            "permission": permission.value,
            "update-all-stream-permissions": unshare_all_pipelines,
        }
        if exclude_pipeline_ids is not None:
            params["exclude-stream-ids"] = exclude_pipeline_ids

        rsp = self.http_client.make_authed_request(
            path=f"{self.service_uri}/data-products/{str(data_product_id)}/shares",
            method=HttpMethod.DELETE,
            params=params,
        )

        if isinstance(rsp, DtlError):
            return rsp

        return UnsharePermission(
            ObjectType.DataProduct, target_id, target_type, permission
        )



import abc
from typing import Dict, Optional, List, Union
from uuid import UUID

from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list


class Path(object):
    def __init__(self, prev_full_path: Optional["Path"]):
        self._prev_full_path = prev_full_path
        self.cur_full_path = list()
        if prev_full_path is not None:
            self.extend(prev_full_path)

    def append(self, path_val):
        self.cur_full_path.append(path_val)

    def extend(self, prev_full_path):
        self.cur_full_path.extend(prev_full_path.cur_full_path)

    def __hash__(self):
        return hash((self._prev_full_path, str(self.cur_full_path)))

    def __eq__(self, other):
        return (self._prev_full_path, str(self.cur_full_path)) == (
            other._prev_full_path,
            str(other.cur_full_path),
        )

    def __str__(self):
        return "/".join(self.cur_full_path)


class Node:
    """
    Describes schema node
    """

    def __init__(self, id: UUID, label: str):
        """
        Node object

        :param id: schema node id
        """

        self.id = id
        self.label = label


class NodeDef(object):
    """
    Represents a pointer to node of a graph.
    """

    def __init__(self, node, siblings: List["Node"], prev_node_id: Optional[str]):
        """
        :param node: Node information for the current node.
        :param siblings: W Node information for current node's connected nodes.
        :param full_path: Full path of the current node.
        """
        self.node = node
        self.siblings = siblings
        self.full_path = Path(None)
        self.prev_node_id = prev_node_id


class Graph:
    """
    Graph representation.
    """

    def __init__(
        self,
        id_to_node: Dict[str, NodeDef],
        path_to_node: Dict[Path, NodeDef],
        root: Optional[str],
    ):
        """
        :param id_to_node: represents mapping between the ID of a node and its `NodeDef`, which contains information of that particular node and its connected siblings.
        :param path_to_node. Path to current node from root, joined by "/"
        :param root: ID of the root node.
        """
        self.id_to_node = id_to_node
        self.path_to_node = path_to_node
        self.root = root

    def get_node_at_path(self, paths: List[str]) -> Union[None, Node]:
        def _get_node_at_path(paths: List[str], cur_node: Node):
            if len(paths) == 0:
                return None
            cur_path = paths[0]
            updated_paths = paths.copy()
            updated_paths.pop(0)
            if cur_path != cur_node.label:
                return None
            if len(updated_paths) == 0 and cur_path == cur_node.label:
                return cur_node
            else:
                siblings: List[Node] = self.id_to_node[str(cur_node.id)].siblings
                for sibling in siblings:
                    node_res = _get_node_at_path(updated_paths, sibling)
                    if isinstance(node_res, Node):
                        return node_res
                return None

        if len(paths) == 0:
            return None
        root_node = self.id_to_node[self.root].node
        return _get_node_at_path(paths, root_node)

    @staticmethod
    def _from_payload(ads_json: dict, node_decoder) -> Union[DtlError, "Graph"]:
        def _parse_list_schema_nodes(json: dict) -> Union[DtlError, List[Node]]:
            schema_nodes = json.get("nodes")
            if schema_nodes is not None:
                schema_nodes = _parse_list(node_decoder)(schema_nodes)
            return schema_nodes

        def _parse_list_edges(json: dict) -> Union[DtlError, List[Edge]]:
            edges = json.get("edges")
            if edges is not None:
                edges = _parse_list(Edge._from_payload)(edges)
            return edges

        def _parse_root_id(json: dict) -> Union[DtlError, str]:
            root = json.get("root")
            if root is not None:
                root = str(root)
            return root

        def _build_full_path(graph: Graph) -> Graph:
            """
            Create full path for each node in a graph by iterating through all nodes in a graph starting from root. Modification is performed in a mutable-way
            :param graph: The graph which has the nodes that will need their full-path to be built for.
            Graph traversal method is BFS.
            :return: The graph.
            """

            def get_next_nodes(node: Node, visited_nodes: set):
                """
                Get the next level nodes, and exclude any previously visited nodes
                :param node: Parent node from which its connected nodes we will visit next
                :return:
                """
                next_nodes = []
                for sibling in node.siblings:
                    if sibling.id not in visited_nodes:
                        next_nodes.append(sibling)
                        visited_nodes.add(sibling.id)
                return next_nodes

            root_node: Node = graph.id_to_node[graph.root].node
            next_nodes = [root_node]
            visited_nodes = {root_node.id}
            while len(next_nodes) > 0:
                to_process_nodes = next_nodes
                new_next_nodes = []
                while len(to_process_nodes) > 0:
                    cur_node_id = to_process_nodes.pop(0).id
                    cur_node: NodeDef = graph.id_to_node[str(cur_node_id)]
                    prev_node_id = cur_node.prev_node_id
                    if prev_node_id is not None:
                        prev_full_path = graph.id_to_node[prev_node_id].full_path
                        cur_node.full_path.extend(prev_full_path)
                    cur_node.full_path.append(cur_node.node.label)
                    graph.path_to_node[cur_node.full_path] = cur_node

                    new_next_nodes.extend(get_next_nodes(cur_node, visited_nodes))
                next_nodes = new_next_nodes
            return graph

        graph = Graph(dict(), dict(), None)
        if ads_json is None:
            return graph
        parsed_root_id = _parse_root_id(ads_json)
        if parsed_root_id is None:
            return graph
        else:
            graph.root = parsed_root_id
        parsed_schema_nodes = _parse_list_schema_nodes(ads_json)
        if isinstance(parsed_schema_nodes, DtlError):
            return parsed_schema_nodes
        parsed_edges = _parse_list_edges(ads_json)
        if isinstance(parsed_edges, DtlError):
            return parsed_edges

        for node in parsed_schema_nodes:
            graph.id_to_node[str(node.id)] = NodeDef(node, list(), None)

        for edge in parsed_edges:
            source_node_id = str(edge.source)
            target_node_id = str(edge.target)
            target_node_def: NodeDef = graph.id_to_node[target_node_id]
            existing_source_node_def: NodeDef = graph.id_to_node[source_node_id]
            existing_source_node_def.siblings.append(target_node_def.node)
            target_node_def.prev_node_id = existing_source_node_def.node.id

        graph = _build_full_path(graph)

        return graph

    def get_nodes(self) -> List[NodeDef]:
        def _get_nodes(cur_nodes: List[NodeDef]):
            nodes = []
            next_nodes = []
            for node in cur_nodes:
                nodes.append(node)
                for sibling in node.siblings:
                    sibling_node_def = self.id_to_node[sibling.id]
                    next_nodes.append(sibling_node_def)

            if len(next_nodes) > 0:
                next_nodes_def = _get_nodes(next_nodes)
                nodes.extend(next_nodes_def)
            return nodes

        if self.root is None:
            return []
        root_node_def = self.id_to_node[self.root]
        nodes_def = []
        nodes_def.extend(_get_nodes([root_node_def]))
        return nodes_def


class Edge:
    """
    Represents edge between AbstractDataSchema. Currently this class is only used for the purpose of deserialization from backend.
    """

    def __init__(self, source: str, target: str):
        self.source = source
        self.target = target
        """
        :param source: Node ID of source node.
        :param target: Node ID of target node.
        """

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Edge"]:
        source = json.get("source")
        if source is None:
            return DtlError("Edge has to have a 'source' key")

        target = json.get("target")
        if target is None:
            return DtlError("Edge has to have a 'target' key")

        return Edge(source, target)



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Union, Optional, List
from uuid import UUID

from datalogue.dtl_utils import SerializableStringEnum
from datalogue.errors import _enum_parse_error, DtlError


class AuthenticationType(SerializableStringEnum):
    LDAP = "LDAP"
    OpenId = "OpenId"

    @staticmethod
    def authentication_type_from_str(
        string: str,
    ) -> Union[DtlError, "AuthenticationType"]:
        return SerializableStringEnum.from_str(AuthenticationType)(string)

    @staticmethod
    def parse_error(s: str) -> DtlError:
        return DtlError(_enum_parse_error("authentication type", s))


class Authentications(ABC):
    def __init__(
        self, name: str, authentication_type: AuthenticationType, activated: bool = True
    ):
        """
        Builds an Authentication

        :param name: Name of the Authentication
        :param authentication_type: Type of Authentication. Ex: LDAP, OpenId
        :param activated: Is the authentication activated
        """
        self.name = name
        self.provider_type = authentication_type
        self.activated = activated

    def __eq__(self, other: "Authentications"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(name: {self.name}, provider_type: {self.provider_type}, "
            f"activated: {self.activated!r})"
        )

    def _base_payload(self) -> dict:
        base = {
            "name": self.name,
            "providerType": self.provider_type.value,
            "activated": self.activated,
        }

        return base

    @abstractmethod
    def _as_payload(self):
        pass


class AuthenticationsReference:
    def __init__(
        self,
        ref_id: UUID,
        org_id: UUID,
        name: str,
        activated: bool,
        provider_type: AuthenticationType,
        config: Authentications,
        created_at: datetime,
        created_by: UUID,
        updated_at: datetime,
        updated_by: UUID,
        owner: UUID,
    ):
        """
        Represents a reference to an Authentication

        :param ref_id: UUID of the authentication
        :param org_id: UUID of the organization to which this authentication belongs to
        :param name: Name of the Authentication
        :param activated: Activation status of the authentication
        :param provider_type: Type of authentication provider Ex: LDAP, OpenId
        :param config: Provider specific configuration
        :param created_at: Creation timestamp
        :param created_by: UUID of the user who created this authentication
        :param updated_at: Last updated timestamp
        :param updated_by: UUID if the user who last updated this authentication
        :param owner: UUID of the owner of this authentication
        """
        self.id = ref_id
        self.org_id = org_id
        self.name = name
        self.activated = activated
        self.type = provider_type
        self.config = config
        self.created_at = created_at
        self.created_by = created_by
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.owner = owner

    def __eq__(self, other: "AuthenticationType"):
        if isinstance(self, other.__class__):
            return (
                self.id == other.id
                and self.name == other.name
                and self.type == other.type
                and self.config._as_payload() == other.config._as_payload()
                and self.owner == other.owner
            )
        return False

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(id: {self.id}, org_id: {self.org_id}, name: {self.name}, "
            f"activated: {self.activated}, type: {self.type}, config: {self.config}, "
            f"created_at: {self.created_at}, created_by: {self.created_by}, updated_at: {self.updated_at}, "
            f"updated_by: {self.updated_by}, owner: {self.owner})"
        )


# todo - below method is used for listing auth types, but misses out standard fields
def _authentication_ref_from_payload(
    json: dict,
) -> Union[DtlError, AuthenticationsReference]:
    ref_id = json.get("id")
    if not isinstance(ref_id, str):
        return DtlError("An AuthenticationsReference needs a 'id' field")
    else:
        ref_id = UUID(ref_id)

    org_id = json.get("organizationId")
    if isinstance(org_id, str):
        org_id = UUID(org_id)

    name = json.get("name")
    if name is None:
        return DtlError("An AuthenticationsReference needs a 'name' field")

    activated = json.get("activated")

    auth_type = json.get("providerType")
    if not isinstance(auth_type, str):
        return DtlError("An AuthenticationsReference needs a 'providerType' field")
    else:
        auth_type = AuthenticationType.from_str(AuthenticationType)(auth_type).value

    config = json.get("config")
    if isinstance(config, dict):
        if auth_type == AuthenticationType.LDAP.value:
            config = LdapAuthenticationDef.from_payload(config)

    return AuthenticationsReference(
        ref_id, org_id, name, activated, auth_type, config, None, None, None, None, None
    )


class LdapAuthenticationDef(Authentications):
    type_str = AuthenticationType.LDAP

    def __init__(
        self,
        name: str,
        host: str,
        port: int,
        bind_cn: str,
        bind_password: str,
        sub_directory: str,
        group_dn: List[str],
        login_attribute: str = "cn",
        group_member_attribute: str = "memberOf",
        mail_attribute: str = "mail",
        first_name_attribute: Optional[str] = None,
        last_name_attribute: Optional[str] = None,
        display_name_attribute: Optional[str] = "displayName",
    ):
        """
        Definition of a LDAP configuration

        :param name: User friendly name for this Authentication
        :param host: Fully qualified host name. Ex: ldap.domain.com
        :param port: port Ex: 636
        :param bind_cn: Bind CN
        :param bind_password: password
        :param sub_directory: Sub directory. Ex: 'ou=system'
        :param group_dn: The distinguished name(s) referencing one or more groups on the LDAP server, as a list of
            strings
        :param login_attribute: The common name(s) referencing one or more users on the LDAP record
        :param group_member_attribute: login attribute applied as a filter, e.g. "memberOf"
        :param mail_attribute: Can be used to point to the field were Email is stored on the LDAP record, for account
            metadata purposes
        :param first_name_attribute: Can be used to map to the field where first name is stored on the LDAP record, for
            account metadata purposes
        :param last_name_attribute: Can be used to point to the field where last name is stored on the LDAP reord, for
            account metadata purposes
        :param display_name_attribute: Can be used to point to the field where display name is stored on the LDAP
            record, for account metadata purposes
        """
        self.host = host
        self.port = port
        self.bind_cn = bind_cn
        self.bind_password = bind_password
        self.sub_directory = sub_directory
        self.group_dn = group_dn
        self.login_attribute = login_attribute
        self.group_member_attribute = group_member_attribute
        self.mail_attribute = mail_attribute
        self.first_name_attribute = first_name_attribute
        self.last_name_attribute = last_name_attribute
        self.display_name_attribute = display_name_attribute
        super().__init__(name, LdapAuthenticationDef.type_str)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(name: {self.name}, host: {self.host}, port: {self.port}, "
            f"bind_cn: {self.bind_cn}, bind_password: {self.bind_password}, sub_directory: {self.sub_directory}, "
            f"group_dn: {self.group_dn}, login_attribute: {self.login_attribute}, group_member_attribute: "
            f"{self.group_member_attribute}, mail: ${self.mail_attribute}, first_name_attribute: "
            f"{self.first_name_attribute!r}, last_name_attribute: {self.last_name_attribute!r}, "
            f"display_name_attribute: {self.display_name_attribute!r})"
        )

    def _as_payload(self) -> dict:
        payload = self._base_payload()
        config = {
            "host": self.host,
            "port": self.port,
            "bindCn": self.bind_cn,
            "bindPassword": self.bind_password,
            "subDirectory": self.sub_directory,
            "providerType": self.provider_type.value,
            "groupDn": self.group_dn,
            "loginAttribute": self.login_attribute,
            "groupMemberAttribute": self.group_member_attribute,
            "mailAttribute": self.mail_attribute,
        }

        if self.first_name_attribute is not None:
            config["firstNameAttribute"] = self.first_name_attribute

        if self.last_name_attribute is not None:
            config["lastNameAttribute"] = self.last_name_attribute

        if self.display_name_attribute is not None:
            config["displayNameAttribute"] = self.display_name_attribute

        payload["config"] = config
        return payload

    @staticmethod
    def from_payload(json: dict) -> Union[DtlError, "LdapAuthenticationDef"]:
        host = json.get("host")
        if not isinstance(host, str):
            return DtlError("LDAP configuration requires a 'host' field")

        port = json.get("port")
        if not isinstance(port, str):
            return DtlError("LDAP configuration requires a 'port' field")
        else:
            port = int(port)

        bind_cn = json.get("bindCn")
        if not isinstance(bind_cn, str):
            return DtlError("LDAP configuration requires a 'bindCn' field")

        bind_password = json.get("bindPassword")
        if not isinstance(bind_password, str):
            return DtlError("LDAP configuration requires a 'bindPassword' field")

        sub_directory = json.get("subDirectory")
        if not isinstance(sub_directory, str):
            return DtlError("LDAP configuration requires a 'subDirectory' field")

        group_dn = json.get("groupDn")
        if not isinstance(group_dn, List):
            return DtlError("LDAP configuration requires a 'groupDn' field")

        login_attribute = json.get("loginAttribute")
        if not isinstance(login_attribute, str):
            return DtlError("LDAP configuration requires a 'loginAttribute' field")

        group_member_attribute = json.get("groupMemberAttribute")
        if not isinstance(group_member_attribute, str):
            return DtlError("LDAP configuration requires a 'groupMemberAttribute'")

        mail_attribute = json.get("mailAttribute")
        if not isinstance(mail_attribute, str):
            return DtlError("LDAP configuration requires a 'mailAttribute'")

        first_name_attribute = json.get("firstNameAttribute")
        last_name_attribute = json.get("lastNameAttribute")
        display_name_attribute = json.get("displayNameAttribute")

        return LdapAuthenticationDef(
            "",
            host,
            port,
            bind_cn,
            bind_password,
            sub_directory,
            group_dn,
            login_attribute,
            group_member_attribute,
            mail_attribute,
            first_name_attribute,
            last_name_attribute,
            display_name_attribute,
        )



from collections import namedtuple
from datetime import datetime
from typing import Optional, Union, List
from uuid import UUID

from datalogue.errors import DtlError
from datalogue.clients._http import _HttpClient, HttpMethod
from datalogue.models.permission import Permission
from datalogue.models.permission import ObjectType, SharePermission
from datalogue.models.classifier import (
    Classifier,
    ClassificationMethod,
    ClassifierDataPoint,
    ClassifiedDataPoint,
)
from datalogue.models.regex import Regex
from datalogue.models.scope_level import Scope
from datalogue.dtl_utils import _parse_list, is_valid_uuid
from datalogue.clients.ontology import OntologyClient
from datalogue.models.ontology import OntologyNode


class _ClassifierClient:
    """
    Client to interact with the Classifiers
    """

    def __init__(self, http_client: _HttpClient, ontology_client: OntologyClient):
        self.http_client = http_client
        self.service_uri = "/scout"
        self.ontology_client = ontology_client

    def _fill_domain(self, classifier: Classifier, class_ids: List[str]):
        classifier.domain = []
        for class_id in class_ids:
            res = self.ontology_client.get_class(UUID(class_id))
            if isinstance(res, DtlError):
                res = OntologyNode(
                    name="Permission denied",
                    description="Permission denied",
                    id=UUID(class_id),
                )
            classifier.domain.append(res)

    def create(self, classifier: Classifier) -> Union[DtlError, Classifier]:
        """
        Creates the Classifier as specified.

        :param classifier: Classifier definition
        :return: Classifier reference with more information id, owner, description
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/classifiers", HttpMethod.POST, classifier._as_payload()
        )

        if isinstance(res, DtlError):
            return res

        new_classifier = Classifier._from_payload(res)
        self._fill_domain(new_classifier, res.get("classIds"))
        return new_classifier

    def update(
        self,
        id: UUID,
        name: Optional[str] = None,
        description: Optional[str] = None,
        classification_methods: Optional[List[ClassificationMethod]] = None,
        default_class_id: Optional[UUID] = None,
    ) -> Union[DtlError, Classifier]:
        """
        Updates the Classifier as specified.

        :param id: id of classifier to update
        :param name: Updated name of the classifier (Optional)
        :param description: Updated description of the classifier (Optional)
        :param classification_methods: Updated classification_methods of the classifier (Optional)
        :param default_class_id: Updated default_class_id of the classifier. If empty string is provided, the default_class_id is set to None. (Optional)
        :return: Regex reference with more information id, owner, regex test sample statuses
        """

        req_body = {}
        if name is not None:
            req_body["name"] = name
        if description is not None:
            req_body["description"] = description
        if classification_methods is not None:
            req_body["classificationMethods"] = list(
                map(lambda m: m._as_payload(), classification_methods)
            )
        if default_class_id is not None:
            if default_class_id == "" or is_valid_uuid(default_class_id):
                req_body["defaultClassId"] = str(default_class_id)
            else:
                raise DtlError(
                    "default_class_id can either be None, empty string, or in UUID format. Other formats are not allowed."
                )

        res = self.http_client.make_authed_request(
            self.service_uri + "/classifiers/" + str(id), HttpMethod.PUT, req_body
        )

        updated_classifier = Classifier._from_payload(res)
        self._fill_domain(updated_classifier, res.get("classIds"))
        return updated_classifier

    def get(self, id: Union[str, UUID]) -> Union[DtlError, Classifier]:
        """
        Retrieve a classifier by its ID.
        :param id: the id of the classifier to be retrieved locally
        :return: the classifier if successful, or a DtlError if failed
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/classifiers/" + str(id), HttpMethod.GET
        )
        if isinstance(res, DtlError):
            return res

        new_classifier = Classifier._from_payload(res)
        self._fill_domain(new_classifier, res.get("classIds"))
        return new_classifier

    def delete(self, classifier_id: Union[str, UUID]) -> Union[DtlError, bool]:
        """
        Delete a classifier.

        :param classifier_id: id of the classifier to be deleted
        :return: True if successful, or DtlError if failed
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/classifiers/" + str(classifier_id), HttpMethod.DELETE
        )

        if isinstance(res, DtlError):
            return res

        return True

    def share(
        self,
        classifier_id: UUID,
        target_id: UUID,
        target_type: Scope,
        permission: Permission,
    ) -> Union[SharePermission, DtlError]:
        """
        Modify the sharing permission for the given classifier. The change can be performed to user, group or organization with the desired permission (Share, Write or Read)
        In order to share, the user needs to have `Share` permission. Classifiers owners always have `Share` permission for that classifier.

        :param classifier_id: UUID is the id of the classifier that you want to share
        :param target_id: UUID is the id of the User, Group or Organization you want to share with (depending on the target_type param)
        :param target_type: Scope (`Organization`, `Group` or `User`) with whom you want to share the classifier. It can be Organization Group or User.
        :param permission: Permission (`Share`, `Write` or `Read`) the permission you want to grant
        :return: If successful returns the permission you granted specifying the `target_id` (UUID),
        the `target_type` (User, Group or Organization) and the `permission` level (Read, Write or Share) you just granted.
        (i.e. {target_id: str, target_type: str, permission: str}),
        else returns :class:`DtlError`.
        """

        url = f"/classifiers/{str(classifier_id)}/shares?targetType={str(target_type.value)}&targetId={str(target_id)}&permission={str(permission.value)}"

        resp = self.http_client.execute_authed_request(
            self.service_uri + url, HttpMethod.POST
        )

        if isinstance(resp, DtlError):
            return resp
        elif resp.status_code == 200:
            return SharePermission(object_type=ObjectType.Classifier).from_payload(
                resp.json()
            )
        else:
            return DtlError(resp.text)

    def list(
        self, page: int = 1, item_per_page: int = 25
    ) -> Union[DtlError, List[Regex]]:
        """
        List all classifiers.

        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available classifiers if successful, or DtlError if failed
        """
        req_params = {}
        if isinstance(page, int):
            req_params["page"] = page
        if isinstance(item_per_page, int):
            req_params["size"] = item_per_page
        res = self.http_client.make_authed_request(
            self.service_uri + "/classifiers", HttpMethod.GET, params=req_params
        )
        if isinstance(res, DtlError):
            return res

        return _parse_list(Classifier._from_payload)(res)

    def search(
        self,
        by_name: Optional[List[str]] = None,
        by_description: Optional[List[str]] = None,
        by_default_class: Optional[List[str]] = None,
        by_owner: Optional[List[UUID]] = None,
        results_per_page: int = 25,
        page: int = 1,
    ) -> Union[DtlError, List[Classifier]]:
        """
        Search among all backend <object>, using any numbers of queries as
        the search criteria
        Multiple items within a query will be applied with OR logic —
        multiple queries will be applied with AND logic
        The wildcard operators: `?` and `*` can be used as wildcards,
        matching one and any number of additional characters, respectively
        :param by_name: query based on classifier name
        :param by_description: query based on description
        :param by_default_class: query based on default class in domain
        :param by_owner: query based on owner id
        :param results_per_page: number of results to include per page
        :param page: page to display
        :return: a list of classifiers, sorted by query similarity, if
        successful, or DtlError if failed
        """
        req_params = {}
        query = {}
        if isinstance(page, int):
            req_params["page"] = page
        if isinstance(results_per_page, int):
            req_params["size"] = results_per_page

        if isinstance(by_name, list):
            query["name"] = by_name
        if isinstance(by_description, list):
            query["description"] = by_description
        if isinstance(by_default_class, list):
            query["defaultClassId"] = by_default_class
        if isinstance(by_owner, list):
            query["owner"] = by_owner
        req_params["query"] = query

        res = self.http_client.make_authed_request(
            self.service_uri + "/classifiers/search", HttpMethod.POST, body=req_params
        )
        if isinstance(res, DtlError):
            return res

        return _parse_list(Classifier._from_payload)(res)

    def classify_datapoint(
        self,
        classifier_id: Union[str, UUID],
        value: Union[str, float, int, datetime, bool, None],
        field_name: Optional[str] = None,
    ) -> Union[ClassifiedDataPoint, DtlError, None]:
        """
        Test a classifier by classifying a single datapoint with specified value and optional field name.

        :param classifier_id: the classifier to be tested
        :param value: the value of the test datapoint to be classified
        :param field_name: can be used to add a field name to the datapoint.
         In the wild, most datapoints will belong to a field with an associated field name
        :return: the class as an OntologyNode if successful classification and match found,
         None if successful classification but no match found, or DtlError if failed
        """

        datapoint = ClassifierDataPoint(value=value, label=field_name)
        datapoints_list = [datapoint._as_payload()]

        res = self.http_client.make_authed_request(
            self.service_uri + f"/classifiers/{classifier_id}/test",
            HttpMethod.POST,
            body=datapoints_list,
        )
        if isinstance(res, DtlError):
            return res
        classified = res[0]
        if classified is None:
            return None

        class_result = self.ontology_client.get_class(UUID(classified["classId"]))
        if isinstance(class_result, DtlError):
            return class_result
        else:
            return ClassifiedDataPoint(
                test_class=class_result,
                score=classified["score"],
                value=value,
                field_name=field_name,
            )



from abc import ABC, abstractmethod
from typing import List, Union, Optional
from datalogue.errors import DtlError


class DataExtractionRule(ABC):
    def __init__(self, extraction_type: str):
        self.type = extraction_type

    def _base_payload(self) -> dict:
        return {"type": self.type}

    def __eq__(self, other: "DataExtractionRule"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        :return: Dictionary representation of the Data Extraction Rule
        """


class IgnorePrefix(DataExtractionRule):
    def __init__(self, prefix: str):
        super().__init__("IgnorePrefix")
        self.prefix = prefix

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["prefix"] = self.prefix
        return base


class PipelineBehavior(ABC):
    def __init__(self, behaviour_type: str):
        self.type = behaviour_type

    def _base_payload(self) -> dict:
        return {"type": self.type}

    def __eq__(self, other: "PipelineBehavior"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        :return: Dictionary representation of the Pipeline Behaviour
        """


class Write(PipelineBehavior):
    def __init__(self):
        """
        Create the target if nonexistant. Append the output records to the target if existing and if output schema is
        compatible with target schema. If schemas are incompatible, the job will fail.
        """
        PipelineBehavior.__init__(self, "Write")

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def _as_payload(self) -> dict:
        return self._base_payload()

    @staticmethod
    def _from_payload(_: dict) -> Union[DtlError, "Write"]:
        return Write()


class Upsert(PipelineBehavior):
    def __init__(self, key: List[str]):
        """
        Create the target if nonexistant. For existing targets, perform an upsert operation. For every output record,
        search existing records for a match in the specified primary key field. If a match is found, update the
        existing record.  If no match is found, append the new record. If schemas are incompatible, job will fail.
        :param key: List of keys to match on while performing an upsert operation
        """
        PipelineBehavior.__init__(self, "Upsert")
        self.key = key

    def __repr__(self):
        return f"{self.__class__.__name__}(key= {self.key})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["key"] = self.key
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "Upsert"]:
        key = json.get("key")
        if key is None:
            return DtlError("A Upsert pipeline behavior needs a  'key' field")

        return Upsert(key)


class ChangeData(PipelineBehavior):
    def __init__(
        self,
        key: List[str],
        op_code_field: str,
        soft_delete_field: str = "DTL_DELETED",
        extraction_rules: List[DataExtractionRule] = [],
    ):
        """

        :param key: Primary key fields
        :param op_code_field:
        """
        PipelineBehavior.__init__(self, "ChangeData")
        self.key = key
        self.op_code_field = op_code_field
        self.soft_delete_field = soft_delete_field
        self.extraction_rules = extraction_rules

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(key= {self.key}, op_code_field= {self.op_code_field}, "
            f"soft_delete_field= {self.soft_delete_field}, extraction_rules= {self.extraction_rules!r})"
        )

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["key"] = self.key
        base["opCodeField"] = self.op_code_field
        base["softDeleteField"] = self.soft_delete_field
        base["extractionRules"] = self.extraction_rules
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, "ChangeData"]:
        key = json.get("key")
        if key is None:
            return DtlError("A ChangeData pipeline behavior needs a 'key' field")

        op_code_field = json.get("opCodeField")
        if op_code_field is None:
            return DtlError(
                "A ChangeData pipeline behavior needs a 'opCodeField' field"
            )

        soft_delete_field = json.get("softDeleteField")
        if soft_delete_field is None:
            return DtlError(
                "A ChangeData pipeline behavior needs a 'softDeleteField' field"
            )

        extraction_rules = json.get("extractionRules")
        if extraction_rules is None:
            return DtlError(
                "A ChangeData pipeline behavior needs a 'extractionRules' field"
            )

        return ChangeData(key, op_code_field, soft_delete_field, extraction_rules)

    @staticmethod
    def get_change_data_for_informatica_flat(
        keys: List[str], soft_delete_field: str = "DTL_DELETED"
    ) -> "ChangeData":
        return ChangeData(
            keys,
            "INFA_OP_TYPE",
            soft_delete_field,
            [IgnorePrefix("INFA_"), IgnorePrefix("DTL__")],
        )

    @staticmethod
    def get_change_data_for_informatica_nested(
        keys: List[str], soft_delete_field: str = "DTL_DELETED"
    ) -> "ChangeData":
        # to be implemented
        pass

    @staticmethod
    def get_change_data_for_informatica_generic(
        keys: List[str], soft_delete_field: str = "DTL_DELETED"
    ) -> "ChangeData":
        # to be implemented
        pass

    @staticmethod
    def get_change_data_for_informatica_wrapped(
        keys: List[str], soft_delete_field: str = "DTL_DELETE"
    ) -> "ChangeData":
        # to be implemented
        pass


_behavior_parsing_map = dict(
    [
        ("Write", Write._from_payload),
        ("Upsert", Upsert._from_payload),
        ("ChangeData", ChangeData._from_payload),
    ]
)


def _pipeline_behavior_from_payload(json: dict) -> Union[DtlError, PipelineBehavior]:
    type_field = json.get("type")
    if type_field is None:
        return DtlError("The PipelineBehavior should have a field 'type'")

    parsing_function = _behavior_parsing_map.get(type_field)
    if parsing_function is None:
        return DtlError(
            "Looks like '%s' pipeline behavior is not handled by the SDK" % type_field
        )

    return parsing_function(json)


class SourceBehavior(ABC):
    def __init__(self, behaviour_type: str):
        self.type = behaviour_type

    def _base_payload(self) -> dict:
        return {"type": self.type}

    def __eq__(self, other: "SourceBehavior"):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        :return: Dictionary representation of the Source Behaviour
        """


class Read(SourceBehavior):
    def __init__(self):
        SourceBehavior.__init__(self, "Read")

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def _as_payload(self) -> dict:
        return self._base_payload()

    @classmethod
    def _from_payload(cls, _: dict) -> Union[DtlError, "Read"]:
        return cls()


class ChangeDataCapture(SourceBehavior):
    def __init__(self, primary_keys: List[str]):
        SourceBehavior.__init__(self, "ChangeData")
        self.primary_keys = primary_keys

    def __repr__(self):
        return f"{self.__class__.__name__}(primary_keys = {self.primary_keys})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["primaryKey"] = self.primary_keys
        return base

    @classmethod
    def _from_payload(cls, json: dict) -> Union[DtlError, "ChangeDataCapture"]:
        primary_keys = json.get("primaryKey")
        if primary_keys is None:
            return DtlError(
                "A ChangeDataCapture source behavior needs a 'primaryKeys' field"
            )
        return cls(primary_keys)


class ResilientRead(SourceBehavior):
    def __init__(self, column_names: List[str]):
        SourceBehavior.__init__(self, "ResilientRead")
        self.column_names = column_names

    def __repr__(self):
        return f"{self.__class__.__name__}(column_names = {self.column_names})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["columnNames"] = self.column_names
        return base

    @classmethod
    def _from_payload(cls, json: dict) -> Union[DtlError, "ResilientRead"]:
        column_names = json.get("columnNames")
        if column_names is None:
            return DtlError(
                "A ResilientRead source behavior needs a 'columnNames' field"
            )
        return cls(column_names)


_source_behavior_map = dict(
    [("Read", Read._from_payload),
     ("ChangeData", ChangeDataCapture._from_payload),
     ("ResilientRead", ResilientRead._from_payload)]
)


def _source_behavior_from_payload(json: dict) -> Union[DtlError, SourceBehavior]:
    type_field = json.get("type")
    if type_field is None:
        return DtlError("The SourceBehavior should have a field 'type'")

    parsing_function = _source_behavior_map.get(type_field)
    if parsing_function is None:
        return DtlError(
            "Looks like %s source behavior is not handled by the SDK" % type_field
        )

    return parsing_function(json)



from enum import Enum
from typing import Optional, List, Union

from datalogue.errors import DtlError, _enum_parse_error
from datalogue.dtl_utils import (
    SerializableStringEnum,
    exists,
    _parse_datetime,
    _parse_uuid,
    _parse_list,
    map_option,
)
from uuid import UUID
from datetime import datetime


class SearchOrder(Enum):
    created_at_asc = ("CreatedAt", "Asc")
    created_at_desc = ("CreatedAt", "Desc")
    updated_at_desc = ("UpdatedAt", "Desc")
    updated_at_asc = ("UpdatedAt", "Asc")


class RegexStatus(SerializableStringEnum):
    NoMatch = "NoMatch"
    PerfectMatch = "PerfectMatch"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("regex status", s))

    @staticmethod
    def from_str(string: str) -> Union[DtlError, "RegexStatus"]:
        return SerializableStringEnum.from_str(RegexStatus)(string)


class RegexTestSample:
    """
    Representation for test samples for Regex. It contains a text value and
    optionally a status to denote whether it's a Match or NoMatch
    """

    def __init__(self, text: str, status: Optional[RegexStatus] = None):
        self.text = text
        self.status = status

    def __repr__(self):
        return "\n  RegexTestSample(" f"text= {self.text!r}, " f"status= {self.status})"

    def __eq__(self, other):
        return (self.text == other.text) and (self.status == other.status)

    @staticmethod
    def _as_payload(sample: "RegexTestSample") -> Union[DtlError, dict]:
        payload = {"text": sample.text}
        if sample.status is not None:
            payload["status"] = sample.status.value
        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "Regex"]:
        text = payload.get("text")
        if text is None:
            return DtlError("'text' is missing! Invalid json for RegexTestSample")
        status = map_option(payload.get("status"), RegexStatus.from_str)
        return RegexTestSample(text, status)


class Regex:
    """
    Representation for regexes.
    """

    def __init__(
        self,
        name: str,
        pattern: str,
        description: str = "",
        test_data: Optional[Union[List[str], List[RegexTestSample]]] = [],
        id: Optional[UUID] = None,
        owner: Optional[UUID] = None,
        editors: Optional[List[UUID]] = None,
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
        updated_by: Optional[UUID] = None,
    ):
        self.id = id
        self.name = name
        self.pattern = pattern
        self.description = description
        if test_data is not None and all(
            isinstance(test_datum_text, str) for test_datum_text in test_data
        ):
            self.test_data = list(
                map(lambda test_datum: RegexTestSample(test_datum), test_data)
            )
        elif test_data is not None and all(
            isinstance(test_datum_text, RegexTestSample)
            for test_datum_text in test_data
        ):
            self.test_data = test_data
        elif test_data is None:
            self.test_data = test_data
        else:
            raise DtlError(
                "Please provide either list of strings or list of RegexTestSample for test_data parameters."
            )
        self.owner = owner
        self.editors = editors
        self.created_at = created_at
        self.updated_at = updated_at
        self.updated_by = updated_by

    def __eq__(self, other):
        return (
            (self.name == other.name)
            and (self.pattern == other.pattern)
            and (self.description == other.description)
            and (self.test_data == other.test_data)
            and (self.owner == other.owner)
            and (self.editors == other.editors)
            and (self.created_at == other.created_at)
            and (self.updated_at == other.updated_at)
            and (self.updated_by == other.updated_by)
        )

    def __repr__(self):
        return (
            "Regex("
            f"id= {self.id!r}, "
            f"name= {self.name!r}, "
            f"pattern= {self.pattern!r},  "
            f"description= {self.description!r}, "
            f"test_data= {self.test_data})"
        )

    @staticmethod
    def _as_payload(regex: "Regex") -> Union[DtlError, dict]:
        payload = {
            "name": regex.name,
            "pattern": regex.pattern,
            "description": regex.description,
            "testData": list(
                map(lambda t: RegexTestSample._as_payload(t), regex.test_data)
            ),
        }

        if regex.id is not None:
            payload["id"] = regex.id

        if regex.owner is not None:
            payload["owner"] = regex.owner

        if regex.editors is not None:
            payload["editors"] = regex.editors

        if regex.created_at is not None:
            payload["createdAt"] = regex.created_at

        if regex.updated_at is not None:
            payload["updatedAt"] = regex.updated_at

        if regex.updated_by is not None:
            payload["lastUpdateBy"] = regex.updated_by

        return payload

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, "Regex"]:
        name = payload.get("name")
        if name is None or not isinstance(name, str):
            return DtlError("'name' is missing or not a string")

        pattern = payload.get("pattern")
        if pattern is None or not isinstance(pattern, str):
            return DtlError("'pattern' is missing or not a string")

        description = payload.get("description")
        if description is None or not isinstance(description, str):
            return DtlError("'description' is missing or not a string")

        test_data_jsons = payload.get("testData")
        if test_data_jsons is None or not isinstance(test_data_jsons, List):
            return DtlError("'testData' is missing or not a List")

        test_data = _parse_list(RegexTestSample._from_payload)(test_data_jsons)
        if isinstance(test_data, DtlError):
            return test_data

        id = map_option(payload.get("id"), _parse_uuid)
        owner = map_option(payload.get("owner"), _parse_uuid)

        editors = payload.get("editors")
        if editors is not None:
            editors = _parse_list(_parse_uuid)(editors)
            if isinstance(editors, DtlError):
                return DtlError(
                    f"'editors has at least one invalid value: {payload.get('editors')}",
                    editors,
                )

        created_at = map_option(payload.get("createdAt"), _parse_datetime)
        if isinstance(created_at, DtlError):
            return created_at

        updated_at = map_option(payload.get("updatedAt"), _parse_datetime)
        if isinstance(updated_at, DtlError):
            return updated_at

        updated_by = map_option(payload.get("lastUpdateBy"), _parse_uuid)
        if isinstance(updated_by, DtlError):
            return updated_by

        return Regex(
            name,
            pattern,
            description,
            test_data,
            id,
            owner,
            editors,
            created_at,
            updated_at,
            updated_by,
        )
